/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2009 Bull S.A.S.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.ow2.jonas.tests;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.List;
import java.util.Map;

import javax.management.Attribute;
import javax.management.MBeanServerConnection;
import javax.management.ObjectName;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;

import org.apache.commons.io.IOUtils;

import org.testng.Assert;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

/**
 * Tests the JOnAS - CARGO plug-in. This test is expected to work with JOnAS 4
 * and JOnAS 5 without any issues.
 *
 * @author S. Ali Tokmen
 * @author Florent Benoit junit->TestNG and Connection leaks checks
 */
public class CargoTest {

    private static final int TIMEOUT = 10;

    private int port;
    private int rmiPort;

    @BeforeClass
    public void setUp() {
        String webcontainerPort = System.getProperty("webcontainer.port");
        Assert.assertNotNull(webcontainerPort, "webcontainer.port not set!");
        this.port = Integer.parseInt(webcontainerPort);
        String carolPort = System.getProperty("carol.port");
        Assert.assertNotNull(carolPort, "carolPort not set!");
        this.rmiPort = Integer.parseInt(carolPort);
    }

    @Test
    public void testCargo() throws Exception {
        final URL url = new URL("http://localhost:" + port + "/cargocpc/index.html");
        final String expected = "Cargo Ping Component used to verify if the container is started.";

        String content = null;
        final long stopTime = System.currentTimeMillis() + CargoTest.TIMEOUT * 1000;
        while (System.currentTimeMillis() < stopTime) {
            Thread.sleep(1000);
            content = this.getContentAsString(url);
            if (content.contains(expected)) {
                break;
            }
        }

        Assert.assertTrue(content.contains(expected), content + " does not contain " + expected);
    }

    /**
     * Test that the connection leak closing mechanism is working
     * @throws Exception
     */
    @Test
    public void testConnectionLeak() throws Exception {
        // Connect to the JMX server
        MBeanServerConnection mBeanServerConnection = getMBeanServerConnection();

        int CONNECTION_LEAKS_REQUEST = 3;

        ObjectName datasourceON =
            new ObjectName("jonas:j2eeType=JCAConnectionFactory,name=jdbc_1,JCAResource=jdbc_1,J2EEServer=jonas");
        ObjectName jndiJDBCInterceptorON = new ObjectName("jonas:type=service,name=jndi-interceptors,interceptor="
            + "org.ow2.jonas.jndi.interceptors.impl.datasource.DataSourceLeakDetectorContextInterceptor");

        // Check there is no busy connections
        Integer currentBusy = (Integer) mBeanServerConnection.getAttribute(datasourceON, "currentBusy");
        Assert.assertNotNull(currentBusy);
        Assert.assertEquals(currentBusy.intValue(), 0);

        // Do some connection leaks
        for (int i = 0; i < CONNECTION_LEAKS_REQUEST; i++) {
            hitConnectionLeakServlet();
        }

        // Check busy are still to 0 (automatic close of connection leak is working)
        // Wait a little
        Thread.sleep(2000);
        currentBusy = (Integer) mBeanServerConnection.getAttribute(datasourceON, "currentBusy");
        Assert.assertEquals(currentBusy.intValue(), 0);

        // Disable automatic close of connection leaks
        Attribute attribute = new Attribute("forceClose", false);
        mBeanServerConnection.setAttribute(jndiJDBCInterceptorON, attribute);

        // Do some connection leaks
        for (int i = 0; i < CONNECTION_LEAKS_REQUEST; i++) {
            hitConnectionLeakServlet();
        }

        // Check that now we have busy connections
        // Wait a little
        Thread.sleep(2000);
        currentBusy = (Integer) mBeanServerConnection.getAttribute(datasourceON, "currentBusy");
        Assert.assertEquals(currentBusy.intValue(), CONNECTION_LEAKS_REQUEST);
    }

    @Test
    public void deactivateDevelopmentMode() throws Exception {
        // Connect to the JMX server
        MBeanServerConnection mBeanServerConnection = getMBeanServerConnection();

        ObjectName depmonitor = new ObjectName("jonas:type=service,name=depmonitor");
        mBeanServerConnection.setAttribute(depmonitor, new Attribute("developmentMode", Boolean.FALSE));
    }

    @Test(dependsOnMethods = "deactivateDevelopmentMode")
    public void testBugJonas96() throws Exception {
        checkIfFileAppearsAsDeployed("dummy.ear", "Ear");
        checkIfFileAppearsAsDeployed("dummy.jar", "Jar");
        checkIfFileAppearsAsDeployed("dummy.rar", "Rar");
        checkIfFileAppearsAsDeployed("dummy.war", "War");
        checkIfFileAppearsAsDeployed("dummy.xml", "File");
    }

    @Test(dependsOnMethods = "deactivateDevelopmentMode")
    public void testJonas292() throws Exception {
        // Connect to the JMX server
        MBeanServerConnection mBeanServerConnection = getMBeanServerConnection();

        final ObjectName j2eeServer = new ObjectName("jonas:j2eeType=J2EEServer,name=jonas");

        final String jonasBase = (String) mBeanServerConnection.getAttribute(j2eeServer, "jonasBase");
        Assert.assertNotNull(jonasBase, "jonasBase is null");
        final String fileName = "jonas-292.jar";
        final File targetFile = new File(jonasBase, "deploy/" + fileName);
        final Object[] opParams = {targetFile.getAbsolutePath()};
        final String[] opSignature = {"java.lang.String"};

        this.checkDeployableStatus(j2eeServer, "deployableJars", fileName, false);
        this.checkDeployableStatus(j2eeServer, "deployableFiles", fileName, false);
        this.checkDeployableStatus(j2eeServer, "deployedJars", fileName, false);
        this.checkDeployableStatus(j2eeServer, "deployedFiles", fileName, false);

        try {
            final InputStream fileContents = this.getClass().getClassLoader().getResourceAsStream(fileName);
            Assert.assertNotNull(fileContents, "Cannot find " + fileName);

            FileOutputStream targetStream = new FileOutputStream(targetFile);
            IOUtils.copy(fileContents, targetStream);

            targetStream.close();
            targetStream = null;
        } finally {
            System.gc();
        }

        try {
            this.checkDeployableStatus(j2eeServer, "deployableJars", fileName, true);
            this.checkDeployableStatus(j2eeServer, "deployableFiles", fileName, true);
            this.checkDeployableStatus(j2eeServer, "deployedJars", fileName, false);
            this.checkDeployableStatus(j2eeServer, "deployedFiles", fileName, false);

            mBeanServerConnection.invoke(j2eeServer, "deploy", opParams, opSignature);

            this.checkDeployableStatus(j2eeServer, "deployableJars", fileName, false);
            this.checkDeployableStatus(j2eeServer, "deployableFiles", fileName, false);
            this.checkDeployableStatus(j2eeServer, "deployedJars", fileName, true);
            this.checkDeployableStatus(j2eeServer, "deployedFiles", fileName, true);

            mBeanServerConnection.invoke(j2eeServer, "undeploy", opParams, opSignature);

            this.checkDeployableStatus(j2eeServer, "deployableJars", fileName, true);
            this.checkDeployableStatus(j2eeServer, "deployableFiles", fileName, true);
            this.checkDeployableStatus(j2eeServer, "deployedJars", fileName, false);
            this.checkDeployableStatus(j2eeServer, "deployedFiles", fileName, false);
        } finally {
            Assert.assertTrue(targetFile.delete(), "File " + targetFile + " cannot be deleted");
        }
    }

    private void checkDeployableStatus(final ObjectName j2eeServer, final String attribute, final String fileName,
        final boolean status) throws Exception {

        // Connect to the JMX server
        MBeanServerConnection mBeanServerConnection = getMBeanServerConnection();

        boolean found = false;
        List<String> files = (List<String>) mBeanServerConnection.getAttribute(j2eeServer, attribute);
        for (String file : files) {
            if (file.contains(fileName)) {
                found = true;
                break;
            }
        }

        Assert.assertEquals(found, status, "File " + fileName + " is " + (status ? "not " : "") + "in the " + attribute
            + " list: " + files);
    }

    /**
     * @return MBean Server connection to the remote server
     * @throws IOException
     */
    protected MBeanServerConnection getMBeanServerConnection() throws IOException {
        // The address of the connector server
        String url = "service:jmx:rmi://localhost/jndi/rmi://localhost:" + rmiPort + "/jrmpconnector_jonas";
        JMXServiceURL address = new JMXServiceURL(url);

        // The environment map, null in this case
        Map<String, ?> environment = null;

        // Connector
        JMXConnector jmxConnector = JMXConnectorFactory.connect(address, environment);

        // Get connection
        return jmxConnector.getMBeanServerConnection();
    }

    /**
     * Make a request and check that the result is OK
     */
    protected void hitConnectionLeakServlet() throws Exception {
        // Connect on the servlet that is making connection leaks
        final URL url = new URL("http://localhost:" + port + "/connection-leaks/");
        final String expected = "Example OK";

        String content = null;
        final long stopTime = System.currentTimeMillis() + CargoTest.TIMEOUT * 1000;
        while (System.currentTimeMillis() < stopTime) {
            Thread.sleep(1000);
            content = this.getContentAsString(url);
            if (content.contains(expected)) {
                break;
            }
        }
        // Check expected value
        Assert.assertTrue(content.contains(expected), content + " does not contain " + expected);
    }

    private void checkIfFileAppearsAsDeployed(final String fileName, final String type) throws Exception {
        // Connect to the JMX server
        List<String> files;
        MBeanServerConnection jmx = getMBeanServerConnection();
        ObjectName j2eeServer = new ObjectName("jonas:j2eeType=J2EEServer,name=jonas");

        files = (List<String>) jmx.getAttribute(j2eeServer, "deployable" + type + "s");
        for (String file : files) {
            Assert.assertFalse(file.contains(fileName), "File " + fileName + " already exists: " + file);
        }
        files = (List<String>) jmx.getAttribute(j2eeServer, "deployableFiles");
        for (String file : files) {
            Assert.assertFalse(file.contains(fileName), "File " + fileName + " already exists: " + file);
        }

        String jonasBase = (String) jmx.getAttribute(j2eeServer, "jonasBase");
        Assert.assertNotNull(jonasBase, "jonasBase is null");
        File testFile = new File(jonasBase, "deploy/" + fileName);
        testFile.createNewFile();

        boolean found = false;
        files = (List<String>) jmx.getAttribute(j2eeServer, "deployable" + type + "s");
        for (String file : files) {
            if (file.contains(file)) {
                found = true;
                break;
            }
        }
        Assert.assertTrue(found, "File " + fileName + " doesn't appear in the deployable" + type + "s list");
        files = (List<String>) jmx.getAttribute(j2eeServer, "deployableFiles");
        for (String file : files) {
            if (file.contains(file)) {
                return;
            }
        }
        Assert.fail("File " + fileName + " doesn't appear in the deployableFiles list");
    }

    /**
     * Get the URL content as a String.
     * @param url the url
     * @return the URL content as a String
     * @throws IOException if an I/O exception occurs.
     */
    private String getContentAsString(final URL url) throws IOException {
        final InputStream inStream = url.openStream();
        final ByteArrayOutputStream bos = new ByteArrayOutputStream();
        try {
            int c;
            while ((c = inStream.read()) != -1) {
                bos.write(c);
            }
        } finally {
            inStream.close();
            bos.flush();
        }
        final String urlString = bos.toString();
        return urlString;
    }
}
